(qtpylib.crossed_above(dataframe['rsi'], 70)) &  # Signal: RSI crosses above 70
(dataframe['tema'] > dataframe['bb_middleband']) &  # Guard: tema above BB middle
(dataframe['tema'] < dataframe['tema'].shift(1)) &  # Guard: tema is falling
