Integer(60, 100, name='sell-rsi-value'),
Categorical([True, False], name='sell-rsi-enabled'),
Categorical(['sell-bb_upper',
             'sell-macd_cross_signal',
             'sell-sar_reversal'], name='sell-trigger')
