"order_types": {
    "buy": "limit",
    "sell": "limit",
    "emergencysell": "limit",
    "stoploss": "limit",
    "stoploss_on_exchange": false
},
"exchange": {
    "name": "{{ exchange_name | lower }}",
    "key": "{{ exchange_key }}",
    "secret": "{{ exchange_secret }}",
    "ccxt_config": {"enableRateLimit": true},
    "ccxt_async_config": {
        "enableRateLimit": true,
        "rateLimit": 500
    },
    "pair_whitelist": [
        "ETH/BTC",
        "LTC/BTC",
        "ETC/BTC",
        "DASH/BTC",
        "ZEC/BTC",
        "XLM/BTC",
        "XRP/BTC",
        "TRX/BTC",
        "ADA/BTC",
        "XMR/BTC"
    ],
    "pair_blacklist": [
    ]
}
