"exchange": {
    "name": "{{ exchange_name | lower }}",
    "key": "{{ exchange_key }}",
    "secret": "{{ exchange_secret }}",
    "ccxt_config": {"enableRateLimit": true},
    "ccxt_async_config": {
        "enableRateLimit": true
    },
    "pair_whitelist": [

    ],
    "pair_blacklist": [

    ]
}
